/****************************************************************************
 *
 * program
 *
 * Description
 *
 * Copyright (C) 2007 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include "BPELModelFactory.hh"
#include "InvokeActivity.hh"
#include "SequenceActivity.hh"
#include "FlowActivity.hh"
#include "DummyActivity.hh"
#include "WhileActivity.hh"
#include "RepeatActivity.hh"
#include "Utils.hh"
#include "WSDLFactory.hh"
#include "BPELModel.hh"

#include <cstring>
#include <iostream>

extern "C" {
#include <libxml/tree.h>
#include <libxml/parser.h>
}

using namespace std;

BPELModelFactory::BPELModelFactory( const string& fname ) :
    m_bpel_model( new BPELModel( "Test BPEL Model" ) )
{
    xmlDocPtr m_doc;
    xmlNodePtr m_root;

    xmlSubstituteEntitiesDefault( 1 );
    m_doc = xmlParseFile( fname.c_str() );
    if ( 0 == m_doc ) {
        cerr << "FATAL: An error occurred while parsing '" 
             << fname << "'" << endl;
        abort( );
    }
    m_root = xmlDocGetRootElement( m_doc );

    m_handler[ "sequence" ] = &BPELModelFactory::make_sequence;
    m_handler[ "flow" ] = &BPELModelFactory::make_flow;
    m_handler[ "invoke" ] = &BPELModelFactory::make_invoke;
    m_handler[ "receive" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "if" ] = &BPELModelFactory::make_dummy; // TODO   
    m_handler[ "while" ] = &BPELModelFactory::make_while;    
    m_handler[ "repeatUntil" ] = &BPELModelFactory::make_repeat;
    m_handler[ "pick" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "forEach" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "scope" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "switch" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "assign" ] = &BPELModelFactory::make_dummy; // TODO
    m_handler[ "reply" ] = &BPELModelFactory::make_dummy; // TODO

    make_bpel_model( m_root );
    xmlFreeDoc( m_doc );
}

Activity* BPELModelFactory::make_activity( xmlNodePtr n )
{
    Activity* result = 0;

    if ( !xmlNodeIsText( n ) ) {
        std::map< std::string, xml_handler >::iterator it;
        it = m_handler.find( (char*)n->name );
        if ( it != m_handler.end() ) {
            xml_handler h( it->second );
            result = (this->*h)( n );
        }
    }

    return result;
}

Activity* BPELModelFactory::make_dummy( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making dummy for " << name << endl;
    return new DummyActivity( name );
}

Activity* BPELModelFactory::make_invoke( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making invoke activity for " << name << endl;
    string operation( BPELAttribute( n, "operation" ).as_string() );
    WSDLOperation* op = 0;
    if ( operation.empty() ) {
        cerr << "WARNING: Invoke activity \"" << name << "\" does not "
             << "reference any operation. Ignoring for now." << endl;
    } else {
        op = m_bpel_model->get_operation_by_name( operation );
        if ( 0 == op ) {
            cerr << "FATAL: Invoke activity \"" << name << "\" references "
                 << "an unknown operation \"" << operation << "\"" << endl;
            exit( -1 );
        }
    }
    return new InvokeActivity( name, op );
}

Activity* BPELModelFactory::make_sequence( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making sequence for " << name << endl;
    SequenceActivity* result = new SequenceActivity( name );
    for ( xmlNodePtr tmp = n->children; 0 != tmp; tmp = tmp->next ) {
        Activity* a = make_activity( tmp );
        if ( a ) result->add_activity( a );
    }
    return result;
}

Activity* BPELModelFactory::make_flow( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making flow for " << name << endl;
    FlowActivity* result = new FlowActivity( name );
    for ( xmlNodePtr tmp = n->children; 0 != tmp; tmp = tmp->next ) {
        Activity* a = make_activity( tmp );
        if ( a ) result->add_activity( a );
    }
    return result;
}

Activity* BPELModelFactory::make_while( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making while for " << name << endl;
    WhileActivity* result = new WhileActivity( name );   
    Activity* inner = 0;
    for ( xmlNodePtr tmp = n->children; 0 != tmp && 0 == inner; tmp = tmp->next ) {
        if ( !xmlStrcmp( tmp->name, (xmlChar*)"condition" ) ) {
            double p( BPELAttribute( tmp, "prob" ).as_double() );
            result->set_prob( p );
        } else {
            inner = make_activity( tmp );
            if ( inner ) result->set_inner( inner );
        }
    }

    return result;
}

Activity* BPELModelFactory::make_repeat( xmlNodePtr n )
{
    string name( BPELAttribute( n, "name" ).as_string() );
    cout << "Making repeat for " << name << endl;
    RepeatActivity* result = new RepeatActivity( name );   
    Activity* inner = 0;
    for ( xmlNodePtr tmp = n->children; 0 != tmp; tmp = tmp->next ) {
        if ( xmlStrcmp( tmp->name, (xmlChar*)"condition" ) ) {
            double p( BPELAttribute( tmp, "prob" ).as_double() );
            result->set_prob( p );
            break; // exit
        } else {
            inner = make_activity( tmp );
            if ( inner ) result->set_inner( inner );
        }
    }

    return result;
}


void BPELModelFactory::make_bpel_model( xmlNodePtr n )
{
    for ( xmlNodePtr tmp = n->children; 0 != tmp; tmp = tmp->next ) {
        if ( !xmlStrcmp( tmp->name, (xmlChar*)"import" ) ) {
            string importType( BPELAttribute( tmp, "importType" ).as_string() );
            if ( 0 == importType.compare( "http://schemas.xmlsoap.org/wsdl/" ) ) {
                string location( BPELAttribute( tmp, "location" ).as_string() );
                // process import
                cout << "Processing import from location " << location << endl;
                WSDLFactory wf( location );
                list< WSDLOperation* > op( wf.get_operations() );
                for ( list< WSDLOperation* >::iterator it = op.begin();
                      op.end() != it; ++it ) {
                    m_bpel_model->add_operation( *it );
                }
            } else {
                cout << "Skipping import of unknown type " << importType << endl;
            }
        } else if ( !xmlStrcmp( tmp->name, (xmlChar*)"workload" ) ) {
            // check the type of workload
            if ( 0 == BPELAttribute( tmp, "type" ).as_string().compare( "open" ) ) {
                // open workload
                double arr_rate( BPELAttribute( tmp, "arrivalrate" ).as_double() );
                cout << "BPEL is an open workload with arrival rate=" << arr_rate << endl;
                m_bpel_model->set_type( BPELModel::open_workload );
                m_bpel_model->set_arrival_rate( arr_rate );
            } else {
                // closed workload
                m_bpel_model->set_type( BPELModel::closed_workload );
                int population( BPELAttribute( tmp, "population" ).as_int() );
                double think_time( BPELAttribute( tmp, "thinktime" ).as_double() );
                cout << "BPEL is a closed workload with population=" << population
                     << " think time=" << think_time << endl;
                m_bpel_model->set_population( population );
                m_bpel_model->set_think_time( think_time );
            }
        } else {
            Activity *a = make_activity( tmp );
            if ( 0 != a ) {
                m_bpel_model->set_root_activity( a );
                cout << "finished" << endl;
                break;
            }
        }
    }
}
