/****************************************************************************
 *
 * ServiceDemandVisitor.cc
 *
 * This file is part of bpel2qnbound
 *
 * Copyright (C) 2007 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/
#include "ServiceDemandVisitor.hh"
#include "InvokeActivity.hh"
#include "SequenceActivity.hh"
#include "FlowActivity.hh"
#include "DummyActivity.hh"
#include "WhileActivity.hh"
#include "RepeatActivity.hh"
#include "WSDLOperation.hh"
#include "BPELModel.hh"
#include <iomanip>
#include <iostream>
#include <fstream>

using namespace std;

ServiceDemandVisitor::ServiceDemandVisitor( ) :
    ActivityVisitor( ),
    m_outer_visit_count( 1.0 ),
    m_D_max( 0.0 ),
    m_D( 0.0 ),
    m_D_ave( 0.0 ),
    m_model( 0 )
{

}

void ServiceDemandVisitor::visit_sequence( SequenceActivity* a )
{
    // cout << "Visiting sequence " << a->get_name() << endl;
    list< Activity* > content( a->get_content() );
    for ( list< Activity* >::const_iterator it = content.begin();
          it != content.end(); ++it ) {
        (*it)->accept( this );
    }
}

void ServiceDemandVisitor::visit_invoke( InvokeActivity* a )
{
    //cout << "Visiting invoke " << a->get_name() << endl;
    WSDLOperation* op = a->get_operation( );
    map< string, double > serv_times( op->get_service_times() );
    map< string, double >::const_iterator it;
    for ( it = serv_times.begin(); it != serv_times.end(); ++it ) {
        m_demand[ it->first ] +=  ( it->second ) * m_outer_visit_count;
        m_visits[ it->first ] += m_outer_visit_count;
    }
}

void ServiceDemandVisitor::visit_flow( FlowActivity* a )
{
    // cout << "Visiting flow " << a->get_name() << endl;
    list< Activity* > content( a->get_content() );
    for ( list< Activity* >::const_iterator it = content.begin();
          it != content.end(); ++it ) {
        (*it)->accept( this );
    }
}

void ServiceDemandVisitor::visit_dummy( DummyActivity* a )
{
    // cout << "Visiting dummy " << a->get_name() << endl;
    // nothing to do
}

void ServiceDemandVisitor::visit_while( WhileActivity* a )
{
    // cout << "Visiting while " << a->get_name() << endl;
    double save_visit_count( m_outer_visit_count );
    double p( a->get_prob() );
    m_outer_visit_count = m_outer_visit_count * p / (1.0 - p );
    a->get_inner()->accept( this );
    m_outer_visit_count = save_visit_count;
}

void ServiceDemandVisitor::visit_repeat( RepeatActivity* a )
{
    // cout << "Visiting repeat " << a->get_name() << endl;
    double save_visit_count( m_outer_visit_count );
    double p( a->get_prob() );
    m_outer_visit_count = m_outer_visit_count / p;
    a->get_inner()->accept( this );
    m_outer_visit_count = save_visit_count;
}

void ServiceDemandVisitor::visit_bpel_model( BPELModel* m )
{
    m_model = m;
    m->get_root_activity()->accept( this );
    // Compute the overall service demand D, and the maximum demand D_max
    map< string, double >::const_iterator it;
    for ( it=m_demand.begin(); it != m_demand.end(); it++ ) {
        if ( it->second > m_D_max ) m_D_max = it->second;
        m_D += it->second;
    }
    m_D_ave = m_D / m_demand.size();
}

#ifdef PIPPO
// void ServiceDemandVisitor::describe_open( const string& fname ) const
// {
//     double lambda = m_model->get_arrival_rate();

//     ofstream f_resp_ab( string( fname+"_resp_ab.gp" ).c_str() );

//     // asymptotic bound for the response time
//     f_resp_ab << "set title \"Asymptotic Bounds for Response Time $R(\\lambda)$\"" << endl
//               << "set key top left" << endl
//               << "set xlabel \"$\\lambda$\"" << endl
//               << "set ylabel \"$R(\\lambda)$\"" << endl
//               << "set xrange [0:]" << endl
//               << "plot [1:] " 
//               << m_D << " title \"$R(\\lambda) \\geq D$\" with lines" << endl;

//     // balanced system bounds for response time
//     ofstream f_resp_bsb( string( fname+"_resp_bsb.gp" ).c_str() );
//     f_resp_bsb << "set title \"Balanced System Bounds for Response Time $R(\\lambda)$\"" << endl
//                << "set key top left" << endl
//                << "set xlabel \"$\\lambda$\"" << endl
//                << "set ylabel \"$R(\\lambda)$\"" << endl
//                << "set xrange [0:]" << endl       
//                << "plot [1:] " << m_D << "/(1-x*" << m_D_ave << ") title \"$R(\\lambda) \\geq D/(1-\\lambda D_\\textit{ave})$\" with lines, "
//                << m_D << "/(1-x*" << m_D_max << ") title \"R(\\lambda) \\leq D/(1-\\lambda D_\\textit{max})$\" with lines" << endl;

//     // asymptotic bounds for throughput    
//     ofstream f_tput_ab( string( fname+"_tput_ab.gp" ).c_str() );
//     f_tput_ab << "set title \"Asymptotic Bounds for Throughput $X(\\lambda)$\"" << endl
//               << "set key top left" << endl
//               << "set xlabel \"$\\lambda$\"" << endl
//               << "set ylabel \"$X(\\lambda)$\"" << endl
//               << "set xrange [0:]" << endl
//               << "plot [1:] " 
//               << 1.0/m_D_max << " title \"$X(\\lambda) \\leq 1/D_\\textit{max}$\" with lines" << endl;

//     // balanced system bounds for throughput    
//     ofstream f_tput_bsb( string( fname+"_tput_ab.gp" ).c_str() );
//     f_tput_bsb << "set title \"Balanced System Bounds Throughput $X(\\lambda)$\"" << endl
//                << "set key top left" << endl
//                << "set xlabel \"$\\lambda$\"" << endl
//                << "set ylabel \"$X(\\lambda)$\"" << endl
//                << "set xrange [0:]" << endl
//                << "plot [1:] " 
//                << 1.0/m_D_max << " title \"$X(\\lambda) \\leq 1/D_\\textit{max}$\" with lines" << endl;
// }

// void ServiceDemandVisitor::describe_closed( const string& fname ) const
// {
//     double Z = m_model->get_think_time();
//     int N = m_model->get_population();
    
//     // asymptotic bound for response time
//     ofstream f_resp_ab( string( fname+"_resp_ab.gp" ).c_str() );
//     f_resp_ab << "set title\"Asymptotic Bound for Response Time $R(N)$\"" << endl
//               << "set key top left" << endl
//               << "set xlabel \"$N$\"" << endl
//               << "set ylabel \"$R(N)$\"" << endl
//               << "set xrange [0:]" << endl
//               << "plot [1:] " 
//               << "x*" << m_D << " title \"$R(N) \\leq ND$\" with lines, "
//               << "( " << m_D << " > x*" << m_D_max << " ? " << m_D << " : " << "x*" << m_D_max << ") title \"$R(N) \\geq \\max(D, N D_\\textit{max})$\" with lines" << endl;
    
//     // balanced system bound for response time
//     ofstream f_resp_bsb( string( fname+"_resp_bsb.gp" ).c_str() );
//     f_resp_bsb << "set title\"Balanced System Bound for Response Time $R(N)$\"" << endl
//                << "set key top left" << endl
//                << "set xlabel \"$N$\"" << endl
//                << "set ylabel \"$R(N)$\"" << endl
//                << "max(a,b) = ( a>b ? a : b )" << endl
//                << "set xrange [0:]" << endl
//                << "plot [1:] " 
//                << "max(x*" << m_D_max << "-" << Z << "," << m_D << "+((x-1)*" << m_D_ave << ")/" << 1.0 + Z/m_D << ") title \"$R(N) \\geq \\cdot$\" with lines, "
//                << "( " << m_D << "+ ( (x-1)*" << m_D_max << ")/(1+" << Z << "/(x*" << m_D << "))) title \"$R(N) \\leq \\cdot$\" with lines" << endl;

//     // asymptotic bounds for throughput
//     ofstream f_tput_ab( string( fname+"_tput_ab.gp" ).c_str() );
//     f_tput_ab << "set title\"Asymptotic Bounds for Throughput $X(N)$\"" << endl
//               << "set key top left" << endl
//               << "set xlabel \"$N$\"" << endl
//               << "set ylabel \"$X(N)$\"" << endl
//               << "min(a,b) = ( a<b ? a : b )" << endl
//               << "set xrange [0:]" << endl
//               << "plot [1:] " 
//               << "x/(x*" << m_D << " + " << Z << ") title \"$X(N) \\geq N/(ND+Z)$\" with lines, "
//               << "min( x/" << m_D+Z << "," << 1.0/m_D_max << ") title \"X(N) \\leq \\min( N/(D+Z), 1/D_\\textit{max})$\" with lines" << endl;    

//     // balanced system bounds for throughput
//     ofstream f_tput_bsb( string( fname+"_tput_bsb.gp" ).c_str() );
//     f_tput_bsb << "set title\"Balanced System Bounds for Throughput $X(N)$\"" << endl
//                << "set key top left" << endl
//                << "set xlabel \"$N$\"" << endl
//                << "set ylabel \"$X(N)$\"" << endl
//                << "min(a,b) = ( a<b ? a : b )" << endl
//                << "set xrange [0:]" << endl
//                << "plot [1:] " 
//                << "x/(" << m_D << "+" << Z << "+( (x-1)*" << m_D_max << ")/(1+" << Z << "/(x*" << m_D << "))) title \"$X(N) \\geq \\cdot$\" with lines, "
//                << "min( x/" << m_D_max << ", x/(" << m_D+Z << "+((x-1)*" << m_D_ave << ")/(1+" << Z/m_D <<"))) title \"$X(N) \\leq \\cdot$\" with lines" << endl;    

// }
#else





void ServiceDemandVisitor::describe_open( const string& fname ) const
{
    double lambda = m_model->get_arrival_rate();

    ofstream f_resp_ab( string( fname+"_resp_ab.gp" ).c_str() );

    // asymptotic bound for the response time
    f_resp_ab << fname <<"_resp_ab_lower(x) =" << m_D << endl;

    // balanced system bounds for response time
    ofstream f_resp_bsb( string( fname+"_resp_bsb.gp" ).c_str() );
    f_resp_bsb << fname << "_resp_bsb_lower(x) =" << m_D << "/(1-x*" << m_D_ave << ")" << endl
               << fname << "_resp_bsb_upper(x) =" << m_D << "/(1-x*" << m_D_max << ")" << endl;

    // asymptotic bounds for throughput    
    ofstream f_tput_ab( string( fname+"_tput_ab.gp" ).c_str() );
    f_tput_ab << fname << "_tput_ab_upper(x) =" << 1.0/m_D_max << endl;

    // balanced system bounds for throughput    
    ofstream f_tput_bsb( string( fname+"_tput_ab.gp" ).c_str() );
    f_tput_bsb << fname << "_tput_bsb_upper(x) = " << 1.0/m_D_max << endl;
}

void ServiceDemandVisitor::describe_closed( const string& fname ) const
{
    double Z = m_model->get_think_time();
    // int N = m_model->get_population();
    
    // asymptotic bound for response time
    ofstream f_resp_ab( string( fname+"_resp_ab.gp" ).c_str() );
    f_resp_ab << fname << "_resp_ab_upper(x) = " << "x*" << m_D << endl
              << fname << "_resp_ab_lower(x) = " << "max( " << m_D << ", x*" << m_D_max << "-" << Z << ")" << endl;
    
    // balanced system bound for response time
    ofstream f_resp_bsb( string( fname+"_resp_bsb.gp" ).c_str() );
    f_resp_bsb << fname << "_resp_bsb_lower(x) = " 
               << "max(x*" << m_D_max << "-" << Z << "," << m_D << "+((x-1)*" << m_D_ave << ")/" << 1.0 + Z/m_D << ")" << endl
               << fname << "_resp_bsb_upper(x) = "
               << m_D << "+ ( (x-1)*" << m_D_max << ")/(1.0+" << Z << "/(x*" << m_D << "))" << endl;

    // asymptotic bounds for throughput
    ofstream f_tput_ab( string( fname+"_tput_ab.gp" ).c_str() );
    f_tput_ab << fname << "_tput_ab_lower(x) =" 
              << "x/(x*" << m_D << " + " << Z << ")" << endl
              << fname << "_tput_ab_upper(x) ="
              << "min( x/" << m_D+Z << "," << 1.0/m_D_max << ")" << endl;

    // balanced system bounds for throughput
    ofstream f_tput_bsb( string( fname+"_tput_bsb.gp" ).c_str() );
    f_tput_bsb << fname << "_tput_bsb_lower(x) ="
               << "x/(" << m_D + Z << "+( (x-1)*" << m_D_max << ")/(1.0+" << Z << "/(x*" << m_D << ")))" << endl
               << fname << "_tput_bsb_upper(x) ="
               << "min( 1.0/" << m_D_max << ", x/(" << m_D+Z << "+((x-1)*" << m_D_ave << ")/(1.0+" << Z/m_D <<")))" << endl;

}
#endif


void ServiceDemandVisitor::describe( const string& fname ) const
{
    cout << "=== ServiceDemandVisitor output follows ===" << endl;
    map< string, double >::const_iterator it;
    for ( it = m_demand.begin(); it != m_demand.end(); ++it ) {
        cout << "Resource " << setw(10) << left << it->first
             << " "
             << "Visits " << setw(10) << setprecision(2) << fixed << m_visits.find( it->first )->second
             << " "
             << "Demand " << setw(10) << setprecision(2) << fixed << it->second
             << endl;
    }
    cout << "=== ServiceDemandVisitor output ends ===" << endl;
    if ( m_model->get_type() == BPELModel::open_workload )
        describe_open( fname );
    else
        describe_closed( fname );
}

